"""
Program name resolution.

Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import os
import sys

from proton.utils.environment import VPNExecutionEnvironment

PROGRAM_NAME = os.path.basename(sys.argv[0]) if sys.argv else "protonvpn"

LOGGING_FILENAME = "vpn-cli"
LOGGING_DIR_PATH = os.path.join(VPNExecutionEnvironment().path_cache, "logs")

HELP_OPTION = "--help"
HELP_OPTION_ABBREVIATED = "-h"
